/**********************************************************************
 * LeechCraft - modular cross-platform feature rich internet client.
 * Copyright (C) 2006-2014  Georg Rudoy
 *
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated by
 * a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************/

#include "lastseen.h"
#include <QSettings>
#include <QCoreApplication>
#include <QIcon>
#include <util/util.h>
#include <util/db/dblock.h>
#include <util/sll/onetimerunner.h>
#include <util/sll/qtutil.h>
#include <util/sll/util.h>
#include <util/sll/delayedexecutor.h>
#include <interfaces/core/icoreproxy.h>
#include <interfaces/core/ipluginsmanager.h>
#include <interfaces/core/iloadprogressreporter.h>
#include <interfaces/azoth/iclentry.h>
#include "ondiskstorage.h"
#include "entrystats.h"

namespace LeechCraft
{
namespace Azoth
{
namespace LastSeen
{
	using LastHash_t = QHash<QString, QDateTime>;
}
}
}

Q_DECLARE_METATYPE (LeechCraft::Azoth::LastSeen::LastHash_t);

namespace LeechCraft
{
namespace Azoth
{
namespace LastSeen
{
	void Plugin::Init (ICoreProxy_ptr proxy)
	{
		Util::InstallTranslator ("azoth_lastseen");

		qRegisterMetaType<LastHash_t> ("LeechCraft::Azoth::LastSeen::LastHash_t");
		qRegisterMetaTypeStreamOperators<LastHash_t> ("LeechCraft::Azoth::LastSeen::LastHash_t");

		Storage_ = std::make_shared<OnDiskStorage> ();

		Migrate (proxy->GetPluginsManager ());
	}

	void Plugin::SecondInit ()
	{
	}

	QByteArray Plugin::GetUniqueID () const
	{
		return "org.LeechCraft.Azoth.LastSeen";
	}

	void Plugin::Release ()
	{
	}

	QString Plugin::GetName () const
	{
		return "Azoth LastSeen";
	}

	QString Plugin::GetInfo () const
	{
		return tr ("Azoth LastSeen displays when a contact has been online and available for the last time.");
	}

	QIcon Plugin::GetIcon () const
	{
		static QIcon icon ("lcicons:/azoth/lastseen/resources/images/lastseen.svg");
		return icon;
	}

	QSet<QByteArray> Plugin::GetPluginClasses () const
	{
		QSet<QByteArray> result;
		result << "org.LeechCraft.Plugins.Azoth.Plugins.IGeneralPlugin";
		return result;
	}

	namespace
	{
		bool IsGoodEntry (QObject *entryObj)
		{
			ICLEntry *entry = qobject_cast<ICLEntry*> (entryObj);
			if (!entry)
			{
				qWarning () << Q_FUNC_INFO
						<< entryObj
						<< "doesn't implement ICLEntry";
				return false;
			}

			return true;
		}
	}

	void Plugin::Migrate (IPluginsManager *manager)
	{
		QSettings settings
		{
			QCoreApplication::organizationName (),
			QCoreApplication::applicationName () + "_Azoth_LastSeen"
		};

		if (settings.allKeys ().isEmpty ())
			return;

		const auto& reporter = manager->CreateLoadProgressReporter (this);

		qDebug () << Q_FUNC_INFO
				<< "gonna migrate";

		const auto& avail = settings.value ("LastAvailable").value<LastHash_t> ();
		const auto& online = settings.value ("LastOnline").value<LastHash_t> ();
		const auto& status = settings.value ("LastStatusChange").value<LastHash_t> ();
		qDebug () << "done reading";

		QHash<QString, EntryStats> stats;

		{
			const auto& proc = reporter->InitiateProcess (tr ("Uniting keys"), 0, 3);

			for (const auto& pair : Util::Stlize (avail))
				stats [pair.first].Available_ = pair.second;

			++*proc;

			for (const auto& pair : Util::Stlize (online))
				stats [pair.first].Online_ = pair.second;

			++*proc;

			for (const auto& pair : Util::Stlize (status))
				stats [pair.first].StatusChange_ = pair.second;

			++*proc;
		}

		qDebug () << "done uniting";

		{
			auto lock = Storage_->BeginTransaction ();

			const auto& proc = reporter->InitiateProcess (tr ("Writing the database"), 0, stats.size ());
			for (const auto& pair : Util::Stlize (stats))
			{
				Storage_->SetEntryStats (pair.first, pair.second);
				++*proc;
			}

			qDebug () << "done writing";

			lock.Good ();

			qDebug () << "done committing";
		}

		settings.clear ();

		qDebug () << "done clearing";
	}

	void Plugin::hookEntryStatusChanged (IHookProxy_ptr, QObject *entryObj, QString variant)
	{
		if (!IsGoodEntry (entryObj))
			return;

		if (variant.isEmpty ())
			return;

		const auto entry = qobject_cast<ICLEntry*> (entryObj);
		const auto& id = entry->GetEntryID ();
		const auto& status = entry->GetStatus ();

		if (!LastState_.contains (id))
		{
			LastState_ [id] = status.State_;
			return;
		}

		const State oldState = LastState_ [id];
		if (oldState == status.State_)
			return;

		LastState_ [id] = status.State_;

		const auto& now = QDateTime::currentDateTime ();

		auto stats = Storage_->GetEntryStats (id).get_value_or ({});
		stats.StatusChange_ = now;

		switch (oldState)
		{
		case SOffline:
		case SProbe:
		case SError:
		case SInvalid:
		case SConnecting:
			return;
		case SOnline:
			stats.Available_ = now;
		default:
			stats.Online_ = now;
			break;
		}

		auto st = Storage_;
		Util::ExecuteLater ([stats, st, id] { st->SetEntryStats (id, stats); });
	}

	void Plugin::hookTooltipBeforeVariants (IHookProxy_ptr proxy, QObject *entryObj)
	{
		if (!IsGoodEntry (entryObj))
			return;

		const auto entry = qobject_cast<ICLEntry*> (entryObj);
		const auto& id = entry->GetEntryID ();

		const auto& maybeStats = Storage_->GetEntryStats (id);
		if (!maybeStats)
			return;

		const auto& stats = *maybeStats;

		QString addition;

		const auto curState = entry->GetStatus ().State_;

		if (curState != SOnline)
		{
			const auto& avail = stats.Available_;
			if (avail.isValid ())
				addition += tr ("Was available: %1")
					.arg (avail.toString ());
		}

		if (curState == SOffline ||
				curState == SError ||
				curState == SInvalid)
		{
			const auto& online = stats.Online_;
			if (online.isValid ())
			{
				if (!addition.isEmpty ())
					addition += "<br/>";
				addition += tr ("Was online: %1")
					.arg (online.toString ());
			}
		}

		const auto& lastChange = stats.StatusChange_;
		if (lastChange.isValid ())
		{
			if (!addition.isEmpty ())
				addition += "<br/>";
			addition += tr ("Last status change: %1")
					.arg (lastChange.toString ());
		}

		if (addition.isEmpty ())
			return;

		const auto& tip = proxy->GetValue ("tooltip").toString ();
		proxy->SetValue ("tooltip", tip + "<br/><br/>" + addition + "<br/>");
	}
}
}
}

LC_EXPORT_PLUGIN (leechcraft_azoth_lastseen, LeechCraft::Azoth::LastSeen::Plugin);
