/**********************************************************************
 * LeechCraft - modular cross-platform feature rich internet client.
 * Copyright (C) 2006-2014  Georg Rudoy
 *
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated by
 * a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************/

#include "quarkmanager.h"
#include <QStandardItemModel>
#include <util/util.h>
#include <util/xdg/itemsfinder.h>
#include <util/xdg/item.h>
#include <util/models/rolenamesmixin.h>
#include "favoritesmanager.h"
#include "itemimageprovider.h"

namespace LeechCraft
{
namespace Launchy
{
	namespace
	{
		class LaunchModel : public Util::RoleNamesMixin<QStandardItemModel>
		{
		public:
			enum Roles
			{
				PermanentID = Qt::UserRole + 1,
				AppName
			};

			LaunchModel (QObject *parent)
			: RoleNamesMixin<QStandardItemModel> (parent)
			{
				QHash<int, QByteArray> roleNames;
				roleNames [Roles::PermanentID] = "permanentID";
				roleNames [Roles::AppName] = "appName";
				setRoleNames (roleNames);
			}
		};
	}

	QuarkManager::QuarkManager (ICoreProxy_ptr proxy, FavoritesManager *favMgr,
			Util::XDG::ItemsFinder *finder, ItemImageProvider *prov, QObject *parent)
	: QObject (parent)
	, Proxy_ (proxy)
	, FavMgr_ (favMgr)
	, Finder_ (finder)
	, ImageProv_ (prov)
	, Model_ (new LaunchModel (this))
	{
		if (Finder_->IsReady ())
			init ();
		else
			connect (Finder_,
					SIGNAL (itemsListChanged ()),
					this,
					SLOT (init ()));

		connect (FavMgr_,
				SIGNAL (favoriteAdded (QString)),
				this,
				SLOT (addItem (QString)));
		connect (FavMgr_,
				SIGNAL (favoriteRemoved (QString)),
				this,
				SLOT (handleItemRemoved (QString)),
				Qt::QueuedConnection);
	}

	QAbstractItemModel* QuarkManager::GetModel () const
	{
		return Model_;
	}

	QStandardItem* QuarkManager::MakeItem (const QString& id) const
	{
		auto item = Finder_->FindItem (id);
		if (!item)
		{
			qWarning () << Q_FUNC_INFO
					<< "item not found"
					<< id;
			return 0;
		}

		ImageProv_->AddItem (item);

		auto modelItem = new QStandardItem;
		modelItem->setData (item->GetPermanentID (), LaunchModel::Roles::PermanentID);

		const auto& curLang = Util::GetLanguage ().toLower ();

		auto descr = item->GetComment (curLang);
		if (descr.isEmpty ())
			descr = item->GetGenericName (curLang);

		auto name = item->GetName (curLang);
		if (!descr.isEmpty ())
			name += " (" + descr + ")";
		modelItem->setData (name, LaunchModel::Roles::AppName);

		return modelItem;
	}

	void QuarkManager::launch (const QString& id)
	{
		auto item = Finder_->FindItem (id);
		if (!item)
		{
			qWarning () << Q_FUNC_INFO
					<< "item not found"
					<< id;
			return;
		}

		item->Execute (Proxy_);
	}

	void QuarkManager::remove (const QString& id)
	{
		FavMgr_->RemoveFavorite (id);
	}

	void QuarkManager::init ()
	{
		if (auto rc = Model_->rowCount ())
			Model_->removeRows (0, rc);

		QList<QStandardItem*> items;
		for (const auto& id : FavMgr_->GetFavorites ())
			if (auto item = MakeItem (id))
				items << item;

		if (!items.isEmpty ())
			Model_->invisibleRootItem ()->appendRows (items);
	}

	void QuarkManager::addItem (const QString& id)
	{
		if (auto item = MakeItem (id))
			Model_->appendRow (item);
	}

	void QuarkManager::handleItemRemoved (const QString& id)
	{
		for (int i = 0, rc = Model_->rowCount (); i < rc; ++i)
			if (Model_->item (i)->data (LaunchModel::Roles::PermanentID) == id)
			{
				Model_->removeRow (i);
				break;
			}
	}
}
}
