/**********************************************************************
 * LeechCraft - modular cross-platform feature rich internet client.
 * Copyright (C) 2006-2014  Georg Rudoy
 *
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated by
 * a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************/

#include "networkmonitor.h"
#include <typeinfo>
#include <QMenu>
#include <QSortFilterProxyModel>
#include <QMainWindow>
#include <QNetworkAccessManager>
#include <util/util.h>
#include <interfaces/core/icoreproxy.h>
#include "requestmodel.h"
#include "headermodel.h"

namespace LeechCraft
{
	namespace Plugins
	{
		namespace NetworkMonitor
		{
			void Plugin::Init (ICoreProxy_ptr proxy)
			{
				Util::InstallTranslator ("networkmonitor");

				NetworkAccessManager_ = proxy->GetNetworkAccessManager ();

				Ui_.setupUi (this);
				connect (Ui_.SearchString_,
						SIGNAL (textChanged (const QString&)),
						this,
						SLOT (filterUpdated ()));
				connect (Ui_.SearchType_,
						SIGNAL (currentIndexChanged (int)),
						this,
						SLOT (filterUpdated ()));

				ProxyModel_ = new QSortFilterProxyModel (this);
				ProxyModel_->setDynamicSortFilter (true);

				Model_ = new RequestModel (this);
				ProxyModel_->setSourceModel (Model_);
				ProxyModel_->setFilterKeyColumn (3);
				Ui_.RequestsView_->setModel (ProxyModel_);
				connect (Ui_.RequestsView_->selectionModel (),
						SIGNAL (currentRowChanged (const QModelIndex&, const QModelIndex&)),
						this,
						SLOT (handleCurrentChanged (const QModelIndex&)));

				Ui_.RequestHeadersView_->setModel (Model_->GetRequestHeadersModel ());
				Ui_.ReplyHeadersView_->setModel (Model_->GetReplyHeadersModel ());

				connect (Ui_.ClearFinished_,
						SIGNAL (toggled (bool)),
						Model_,
						SLOT (setClear (bool)));

				connect (NetworkAccessManager_,
						SIGNAL (requestCreated (QNetworkAccessManager::Operation,
								const QNetworkRequest&, QNetworkReply*)),
						Model_,
						SLOT (handleRequest (QNetworkAccessManager::Operation,
								const QNetworkRequest&, QNetworkReply*)));

				QAction *showAction = new QAction (tr ("Network monitor..."),
						this);
				showAction->setIcon (GetIcon ());
				connect (showAction,
						SIGNAL (triggered ()),
						this,
						SLOT (show ()));
				Actions_.push_back (showAction);
			}

			void Plugin::SecondInit ()
			{
			}

			void Plugin::Release ()
			{
				qDeleteAll (Actions_);
			}

			QByteArray Plugin::GetUniqueID () const
			{
				return "org.LeechCraft.NetworkMonitor";
			}

			QString Plugin::GetName () const
			{
				return "NetworkMonitor";
			}

			QString Plugin::GetInfo () const
			{
				return tr ("Monitors HTTP network requests and responses.");
			}

			QIcon Plugin::GetIcon () const
			{
				static QIcon icon ("lcicons:/resources/images/networkmonitor.svg");
				return icon;
			}

			QList<QAction*> Plugin::GetActions (ActionsEmbedPlace place) const
			{
				if (place == ActionsEmbedPlace::ToolsMenu)
					return Actions_;

				return QList<QAction*> ();
			}

			void Plugin::handleCurrentChanged (const QModelIndex& index)
			{
				Model_->handleCurrentChanged (ProxyModel_->mapToSource (index));
			}

			void Plugin::filterUpdated ()
			{
				QString search = Ui_.SearchString_->text ();
				switch (Ui_.SearchType_->currentIndex ())
				{
					case 0:
						ProxyModel_->setFilterFixedString (search);
						break;
					case 1:
						ProxyModel_->setFilterWildcard (search);
						break;
					case 2:
						ProxyModel_->setFilterRegExp (search);
						break;
					default:
						qWarning () << Q_FUNC_INFO
							<< "unknown search type"
							<< Ui_.SearchType_->currentIndex ()
							<< Ui_.SearchType_->currentText ();
						break;
				}
			}
		};
	};
};

LC_EXPORT_PLUGIN (leechcraft_networkmonitor, LeechCraft::Plugins::NetworkMonitor::Plugin);

