/**********************************************************************
 * LeechCraft - modular cross-platform feature rich internet client.
 * Copyright (C) 2006-2014  Georg Rudoy
 *
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated by
 * a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************/

#pragma once

#include <utility>
#include <QAbstractItemModel>

namespace LeechCraft
{
namespace Util
{
	/** @brief Abstracts away differences between Qt4 and Qt5 in model
	 * DnD support.
	 *
	 * Derive from this class (and parametrize the template by desired
	 * base class like QStandardItemModel or QAbstractItemModel) to
	 * obtain Qt4-style getter and setter functions for supported drag
	 * and drop actions (supportedDragActions(),
	 * setSupportedDragActions(), supportedDropActions(),
	 * setSupportedDropActions()).
	 *
	 * @tparam Model The source model type that you wish to derive from.
	 *
	 * @sa supportedDragActions()
	 * @sa setSupportedDragActions()
	 * @sa supportedDropActions()
	 * @sa setSupportedDropActions()
	 *
	 * @ingroup ModelUtil
	 */
	template<typename Model>
	class DndActionsMixin : public Model
	{
#if QT_VERSION >= 0x050000
		Qt::DropActions Drags_;
		Qt::DropActions Drops_;
#endif
	public:
		/** @brief Constructs the model passing the arguments to the base
		 * constructor.
		 *
		 * @param[in] args The list of parameters to pass to the
		 * constructor of the base \em Model.
		 * @tparam Args The variadic template parameter pack of arguments
		 * for the base \em Model constructor.
		 */
		template<typename... Args>
		DndActionsMixin (Args&&... args)
		: Model { std::forward<Args> (args)... }
#if QT_VERSION >= 0x050000
		, Drags_ { Model::supportedDragActions () }
		, Drops_ { Model::supportedDropActions () }
#endif
		{
		}

#if QT_VERSION >= 0x050000
		Qt::DropActions supportedDragActions () const override
		{
			return Drags_;
		}

		void setSupportedDragActions (Qt::DropActions acts)
		{
			Drags_ = acts;
		}

		Qt::DropActions supportedDropActions () const override
		{
			return Drops_;
		}

		void setSupportedDropActions (Qt::DropActions acts)
		{
			Drops_ = acts;
		}
#endif
	};
}
}
