/**********************************************************************
 * LeechCraft - modular cross-platform feature rich internet client.
 * Copyright (C) 2006-2014  Georg Rudoy
 *
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated by
 * a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************/

#pragma once

#include <QString>
#include "sysconfig.h"

namespace LeechCraft
{
namespace Util
{
namespace SysInfo
{
	/** @brief Returns a string of OS name and version joined together.
	 *
	 * @return The name and version of OS running LeechCraft.
	 */
	UTIL_SYS_API QString GetOSName ();

	/** @brief Describes the OS running LeechCraft.
	 */
	struct OSInfo
	{
		/** @brief The name of the OS, including the distribution.
		 *
		 * Typical values are:
		 * - Gentoo/Linux
		 * - openSUSE 13.1 (Bottle) (x86_64)
		 * - Mac OS X
		 * - Windows
		 *
		 * On non-Linux systems this field typically matches the Flavour_
		 * field. On Linux it also includes the distribution name and
		 * possibly version.
		 *
		 * @sa Flavour_
		 */
		QString Name_;

		/** @brief The full version of the OS.
		 *
		 * This possibly includes the architecture, the OS release and
		 * OS-dependent version components like kernel version on Linux.
		 */
		QString Version_;

		/** @brief The OS flavour, or name of the OS without any
		 * distribution.
		 *
		 * Typical values are:
		 * - Linux
		 * - Mac OS X
		 * - Windows
		 * - FreeBSD
		 *
		 * On non-Linux systems this typically matches the Name_ field.
		 *
		 * @sa Name_
		 */
		QString Flavour_;

		/** @brief Describes the CPU architecture of the OS.
		 *
		 * This describes the architecture of the OS, not the machine
		 * itself. Thus, a 32-bit Linux running on a 64-bit CPU will
		 * still be reported as \em x86 instead of \em x86_64.
		 */
		QString Arch_;

		/** @brief Constructs the OSInfo object.
		 *
		 * Sets both the Name_ and the Flavour_ of the OS to \em name.
		 *
		 * \param[in] arch Initializer for the Arch_ field.
		 * \param[in] name Initializer for the Name_ and Flavour_ fields.
		 * \param[in] version Initializer for the Version_ field.
		 */
		UTIL_SYS_API OSInfo (const QString& arch, const QString& name, const QString& version);

		/** @brief Constructs the OSInfo object.
		 *
		 * \param[in] arch Initializer for the Arch_ field.
		 * \param[in] flavour Initializer for the Flavour_ field.
		 * \param[in] name Initializer for the Name_ field.
		 * \param[in] version Initializer for the Version_ field.
		 */
		UTIL_SYS_API OSInfo (const QString& arch,
				const QString& flavour,
				const QString& name,
				const QString& version);
	};

	/** @brief Returns more precise information about OS name and version.
	 *
	 * @return A structure OSInfo consisting of OS name, version and other
	 * fields.
	 */
	UTIL_SYS_API OSInfo GetOSInfo ();
}
}
}
